import socket
import time
import argparse
import sys
import json

maxWaitTime = 3

class GPIOClient:

    debug = False
    verbose = False
    server_address = '/var/lib/gpioserver/socket'

    def __init__(self, args):
        if 'socket' in args:
            args['server_address'] = args['socket']
        for name in args:
            if hasattr(self, name) and type(getattr(self, name)) in [str,float,int,bool]:
                if self.debug:
                    print('Setting property: %s=%s' % (name, args[name]))
                setattr(self, name, args[name])
        self.connect()

    def connect(self):

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        t0 = time.time()

        while True:
            if self.debug:
                print('connecting to {}'.format(self.server_address))
            try:
                sock.connect(self.server_address)
                break
            except socket.error as msg:
                failmsg = msg
            if time.time() - t0 > maxWaitTime:
                print('Failed to connect to server in %g s' % (maxWaitTime,))
                print(failmsg)
                sys.exit(1)
            time.sleep(0.5)

        if self.debug:
            print('connected to {}'.format(self.server_address))

        self.sockfd = sock.makefile('rw')

    def sendcmd(self, cmd, **kw):
        kw['cmd'] = cmd
        cd = json.dumps(kw)
        self.sockfd.write('%s\n' % (cd,))
        self.sockfd.flush()
        resp = self.sockfd.readline()
        if self.verbose:
            print(resp)
        return resp

    def led(self, ind, val):
        self.sendcmd('led', index=ind, value=val)

    def blink(self, ind, val, ton=0.7, toff=0.3):
        self.sendcmd('led', index=ind, value=val, blink=[ton, toff])

    def get(self, ind):
        self.sendcmd('led', index=ind)

    def setstate(self, v, blink=False):
        self.sendcmd('set', values=v, blink=blink)

    def sequence(self, name, args):
        self.sendcmd('seq', seq=name, clear=args.clear, reset=args.reset)

    def status(self):
        self.sendcmd('status')

    def clear(self):
        self.sendcmd('clear')

    def reset(self):
        self.sendcmd('reset')

    def quitServer(self):
        self.sendcmd('quit')

    def close(self):
        self.sendcmd('bye')

    def version(self):
        self.sendcmd('version')

    def interactive(self):
        while not sys.stdout.closed:
            sys.stdout.write('gcli> ')
            sys.stdout.flush()
            ln = sys.stdin.readline()
            if len(ln) == 0:
                print('input closed, exit')
                break
            if len(ln.strip()) == 0:
                print('empty line, ignore')
                continue
            print('cmd: "%s"' % (ln,))
            self.sockfd.write('%s' % (ln,))
            self.sockfd.flush()
            resp = self.sockfd.readline()
            print(resp)

def parseargs():
    import argparse
    import configparser

    conf_parser = argparse.ArgumentParser(add_help=False)
    conf_parser.add_argument("-c", "--conf_file", default=["/etc/gpioserver.conf", "gpioserver.conf",
                                                           "/etc/gpioclient.conf", "gpioclient.conf"],
                             help="Specify config file", metavar="FILE")
    conf_parser.add_argument("-C", "--conf_section", default="gpioclient",
                             help="Specify config file section", metavar="NAME")
    args, remaining_argv = conf_parser.parse_known_args()

    defaults = {
        'debug': False,
        'socket': '/var/lib/gpioserver/socket',
        'logfile': 'gpioclient.log',
    }
    if args.conf_file:
        config = configparser.ConfigParser(defaults=defaults)
        if False:
            print('defaults', defaults)
            print("Reading configuration file from %s" % (args.conf_file,))
        cread = config.read(args.conf_file)
        if False:
            print("Configuration files read: %s" % (cread,))
        if len(cread):
            if args.conf_section in config:
                defaults = dict(config.items(args.conf_section))
                if False:
                    print('Loading configuration defaults from %s/[%s]' % (cread, args.conf_section))
                    print('defaults', defaults)

    parser = argparse.ArgumentParser(description='GPIO server client',
                                     parents=[conf_parser],
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--debug', action='store_true', help='Debug')
    parser.add_argument('--socket', type=str, help='Server socket name')
    parser.add_argument('--logfile', '-o', type=str, help='Log file name')

    parser.add_argument('--interactive', '-i', action='store_true', help='Start intractive shell')
    parser.add_argument('--get', '-g', action='store_true', help='Get current state')
    parser.add_argument('--blink', '-b', action='store_true', help='Set blinking')
    parser.add_argument('--blink-blue', action='store_true', help='Set blinking')
    parser.add_argument('--blink-red', action='store_true', help='Set blinking')
    parser.add_argument('--blink-cyan', action='store_true', help='Set blinking')
    parser.add_argument('--status', '-S', action='store_true', help='Get all pin states')
    parser.add_argument('--clear', action='store_true', help='Clear all pin states')
    parser.add_argument('--reset', action='store_true', help='Reset all pin states to on')
    parser.add_argument('--setstate', '-s', type=str, help='Set all pin states')
    parser.add_argument('--sequence', type=str, help='Run named sequence')
    parser.add_argument('--server-version', '-V', action='store_true', help='Get server version')

    parser.add_argument('led', type=int, metavar='L', nargs='?', help='LED')
    parser.add_argument('state', type=float, metavar='S', nargs='?', help='Status')

    parser.set_defaults(**defaults)

    args = parser.parse_args(remaining_argv)

    if args.debug:
        print('parsed args', vars(args))
    return args

def run():
    args = parseargs()
    cli = GPIOClient(vars(args))
    cli.verbose = True

    if args.status:
        cli.status()
    elif args.get:
        cli.get(args.led)
    elif args.setstate:
        st = json.loads(args.setstate)
        cli.setstate(st, args.blink)
    elif args.sequence:
        cli.sequence(args.sequence, args)
    elif args.clear:
        cli.clear()
    elif args.reset:
        cli.reset()
    elif args.blink:
        cli.blink(args.led, args.state)
    elif args.blink_blue:
        cli.setstate([0,0,1], [0.75, 0.75])
    elif args.blink_red:
        cli.setstate([1,0,0], [0.75, 0.75])
    elif args.blink_cyan:
        cli.setstate([0,1,1], [0.75, 0.75])
    elif args.interactive:
        cli.interactive()
    elif args.led is not None and args.state is not None:
        cli.led(args.led, args.state)
    elif args.server_version:
        cli.version()
    else:
        print('nothing to do')

def test():
    cli = GPIOClient()
    # cli.led(0, 0)
    # cli.led(1, 0)
    # cli.led(2, 0)
    # time.sleep(1)
    # cli.led(0, 1)
    # time.sleep(1)
    # cli.led(0, 0.66)
    # time.sleep(1)
    # cli.led(0, 0.33)
    # time.sleep(1)
    # cli.led(0, 0)
    # time.sleep(1)
    # cli.led(0, 0.25)
    # cli.led(1, 0.25)
    # time.sleep(1)
    # cli.led(0, 1)
    # cli.led(1, 1)
    # cli.led(2, 0)
    # time.sleep(1)
    # cli.led(0, 1)
    # cli.led(1, 0)
    # cli.led(2, 1)
    # time.sleep(1)
    cli.led(0, 0)
    cli.led(1, 1)
    cli.led(2, 1)
    time.sleep(1)
    cli.get(0)
    cli.get(1)
    cli.get(2)
    cli.status()
    cli.close()

if __name__ == "__main__":
    run()
