import os
import stat
import socket
import gpiozero
import signal
import threading
import json
import time
from . import VERSION

class GPIOServer:
    pins = '[6, 5, 13]'
    conf_file = '/var/lib/gpioserver/socket'
    server_address = '/var/lib/gpioserver/socket'
    stop = False
    connections = {}
    debug = False
    verbose = False
    sequences = {}

    def __init__(self, args={}):
        if 'socket' in args:
            args['server_address'] = args['socket']
        for name in args:
            if hasattr(self, name) and type(getattr(self, name)) in [str,float,int]:
                print('Setting property: %s=%s' % (name, args[name]))
                setattr(self, name, args[name])

    def clean(self, sig, *args):
        print("caught signal %s" %(sig,))
        self.stop = True

    def openSocket(self):
        sockdir = os.path.dirname(self.server_address)
        try:
            os.mkdir(sockdir)
        except OSError:
            if not os.path.exists(sockdir):
                raise

        # Make sure the socket does not already exist
        try:
            os.unlink(self.server_address)
        except OSError:
            if os.path.exists(self.server_address):
                raise

        # Create a UDS socket
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        # Bind the socket to the address
        print('bind to socket %s' % (self.server_address,))
        sock.bind(self.server_address)

        os.chmod(self.server_address, stat.S_IRWXU|stat.S_IRWXG|stat.S_IRWXO)
        
        # Listen for incoming connections
        sock.listen(10)

        # Set timeout
        sock.settimeout(0.5)

        self.sock = sock

    def openPins(self):
        if type(self.pins) == type(''):
            pinConf = json.loads(self.pins)
            print(f'pins loaded: {pinConf}')
            self.pins = [gpiozero.PWMLED(v) for v in pinConf]

    def loadSequence(self, name, sdef):
        self.sequences[name] = sdef

    def loadSequences(self):
        import configparser
        defaults = {'sequence_init': '[{"delay": 0.5, "values": [0, 1, 0]},{"delay": 0.5, "values": [0, 0, 1]}]',
                    'sequence_shutdown': '[{"values": [0, 1, 0], "blink": 1}]'}
        config = configparser.ConfigParser(defaults=defaults)
        cread = config.read(self.conf_file)
        print('read config', cread)
        if config.has_section('gpioserver'):
            items = dict(config.items('gpioserver'))
        else:
            items = dict(config.items('DEFAULT'))
        print(items)
        if 'sequence_init' in items:
            sdef = json.loads(items['sequence_init'])
            self.loadSequence('init', sdef)
        if 'sequence_shutdown' in items:
            sdef = json.loads(items['sequence_shutdown'])
            self.loadSequence('shutdown', sdef)

    def setStatus(self, args):
        blink = False
        if 'blink' in args:
            ton = 0.7
            toff = 0.3
            bset = args['blink']
            if isinstance(bset, list) and len(bset) >= 2:
                ton = bset[0]
                toff = bset[1]
                blink = True
            else:
                blink = bset
        if 'values' in args:
            values = args['values']
            for (i,p) in enumerate(self.pins):
                p.value = values[i]
                if values[i] and blink:
                    p.blink(ton, toff)

    def runSequence(self, name):
        if name not in self.sequences:
            print(f'no sequence {name} defined')
            return
        sdef = self.sequences[name]
        for (i, v) in enumerate(sdef):
            self.setStatus(v)
            if 'delay' in v:
                time.sleep(v['delay'])

    def clear(self):
        for p in self.pins:
            p.value = 0

    def reset(self):
        for p in self.pins:
            p.value = 1

    def handleClosed(self):
        ckeys = list(self.connections.keys())
        for c in ckeys:
            entry = self.connections[c]
            if entry['conn'].closed:
                entry['thread'].join()
                print('Connection id %d is closed, thread joined' % (c,))
                del self.connections[c]
        if self.debug:
            print('Open connections: %s' % (self.connections,))
    
    def server(self):

        for led in self.pins:
            led.on()

        cid = 1

        while not self.stop:
            try:
                connection, client_address = self.sock.accept()
            except socket.timeout as ex:
                self.handleClosed()
                continue
            except BaseException as ex:
                print(ex)
                raise

            connection.setblocking(True)
            confile = connection.makefile('rw')

            print('new connection from %s: %s, %s' %(client_address, connection.getblocking(), connection.gettimeout()))

            thread = threading.Thread(target=self.handle, args=(confile, client_address))
            self.connections[cid] = {'thread': thread, 'conn': confile}
            cid += 1
            thread.start()

            # self.handle(connection.makefile('rw'), client_address)


    def handle(self, conn, client_address):
        while not conn.closed:
            line = conn.readline().strip()
            if len(line) == 0:
                print('connection closed by %s' % (client_address,))
                try:
                    conn.close()
                except:
                    pass
                break

            try:
                args = json.loads(line)
            except BaseException as e:
                print(e)
                conn.write('invalid cmmand data: %s\n' % (e,))
                conn.flush()
                continue

            if self.verbose:
                print('cmd %s from connection %s' % (args, client_address,))

            if 'cmd' not in args:
                print('no cmd in data: %s' %(cmd,))
            cmd = args['cmd']

            if cmd == 'status':
                st = [p.value for p in self.pins]
                conn.write('LEDs: %s\n' % ([p.value for p in self.pins],))

            elif cmd == 'set':
                self.setStatus(args)
                conn.write('LED state set: %s\n' % ([p.value for p in self.pins],))

            elif cmd == 'seq':
                name = args['seq']
                self.runSequence(name)
                conn.write(f'Sequence exec: {name}\n')
                if args['reset']:
                    self.reset()
                if args['clear']:
                    self.clear()

            elif cmd == 'led':
                arg1i = args['index']
                if arg1i < len(self.pins):
                    pin = self.pins[arg1i]
                    print('handling led %d => %d' % (arg1i,pin.pin.number))
                    if 'value' in args and not 'get' in args:
                        arg2f = args['value']
                        print('led %d => %d set %g' % (arg1i,pin.pin.number, arg2f))
                        if arg2f != 0:
                            print('led %d on' % (arg1i,))
                            pin.value = arg2f
                            if 'blink' in args:
                                ton = 0.7
                                toff = 0.3
                                bset = args['blink']
                                if isinstance(bset, list) and len(bset) >= 2:
                                    ton = bset[0]
                                    toff = bset[1]
                                print('LED %d blink %g, %g/%g'% (arg1i, pin.value, ton, toff))
                                pin.blink(ton, toff)
                        else:
                            print('led %d off' % (arg1i,))
                            pin.off()
                        conn.write('ACK: led %d %g\n' % (arg1i, pin.value,))
                    else:
                        # get
                        conn.write('%s\n' % (pin.value,))

            elif cmd == 'reset':
                self.reset()
                conn.write('reset\n')

            elif cmd == 'clear':
                self.clear()
                conn.write('clear\n')

            elif cmd == 'version':
                conn.write('server version: %s\n' % VERSION)

            elif cmd == 'bye' or cmd == 'goodbye':
                conn.write('goodbye\n')
                conn.close()

            elif cmd == 'quit':
                conn.write('server shutting down\n')
                self.stop = True
                conn.close()

            else:
                conn.write('error: unknown command: %s\n' %(line,))

            if not conn.closed:
                try:
                    conn.flush()
                except BaseException as e:
                    print(e)


def parseargs():
  import argparse
  import configparser

  conf_parser = argparse.ArgumentParser(add_help=False)
  conf_parser.add_argument("-c", "--conf_file", default=["/etc/gpioserver.conf", "gpioserver.conf"],
                           help="Specify config file", metavar="FILE")
  conf_parser.add_argument("-C", "--conf_section", default="gpioserver",
                           help="Specify config file section", metavar="NAME")
  args, remaining_argv = conf_parser.parse_known_args()

  defaults = {
    'debug': 0,
    'socket': '/var/lib/gpioserver/socket',
    'logile': 'gpioserver.log',
  }
  if args.conf_file:
    print('defaults', defaults)
    config = configparser.ConfigParser(defaults=defaults)
    print("Reading configuration file from %s" % (args.conf_file,))
    cread = config.read(args.conf_file)
    print("Read configuration files: %s" % (cread,))
    if len(cread):
      defaults = dict(config.items(args.conf_section))
      print('Loading configuration defaults from %s/[%s]' % (cread, args.conf_section))
      print('defaults', defaults)

  parser = argparse.ArgumentParser(description='GPIO server daemon',
                                   parents=[conf_parser],
                                   formatter_class=argparse.RawDescriptionHelpFormatter)

  parser.add_argument('--debug', type=int, help='Debug')
  parser.add_argument('--socket', type=str, help='Server socket name')
  parser.add_argument('--logfile', '-o', type=str, help='Log file name')

  parser.set_defaults(**defaults)

  args = parser.parse_args(remaining_argv)

  print('parsed args', vars(args))
  return args

def run():
    args = parseargs()
    server = GPIOServer(vars(args))

    for sig in (signal.SIGABRT, signal.SIGINT, signal.SIGHUP, signal.SIGTERM):
        signal.signal(sig, lambda *args: server.clean(sig, *args))

    server.openSocket()
    server.openPins()
    server.loadSequences()
    server.runSequence('init')
    server.server()
    print('gpioserver exiting')

if __name__ == "__main__":
    run()
